/**
 * menus managing module
 */
tma.registerModule( menumanager );
menumanager.prototype = new tmaModule();
function menumanager() {

	var self = this;

	// list of menu instances
	var menus = {};

	/**
	 * IMPORTANT: the modules name; equal to directory and script file name
	 */
	this.name = 'menumanager';

	/**
	 * initialize module
	 */
	this.setup = function() {

		// register tmaFeatures
		tma.fn.addFeature( this, 'setPage' );
		tma.fn.addFeature( this, 'addPage' );
		tma.fn.addFeature( this, 'addConfigItem' );

		return this.loadLanguage().then(function() {
			return self.init();
		});

	};

	/**
	 * append tab to menu
	 *
	 * @param {tmaFeature}	f
	 * @param {String}		menuid	menuId
	 * @param {tmaModule}	module	page module
	 */
	this.addPage = function( f, menuid, module ) {

		f.stop();

		if (!menus[menuid]) {
			menus[menuid] = new Menu(menuid);
		}

		/** @returns Menu */
		var menu = menus[menuid];
		menu.addPage( module, module.i18.menu_caption );

	};

	/**
	 * activates a page of a specific module
	 *
	 * @param {tmaFeature}	f
	 * @param {String}		menuid			Menu ID
	 * @param {tmaModule}	module			page module
	 */
	this.setPage = function( f, menuid, module ) {

		if (menus[menuid]) {

			f.stop();

			menus[menuid].setPage(module);
		}

	};

	/**
	 * adds a config item for a module to a menu
	 *
	 * @param {tmaFeature}	f
	 * @param {String}		menuid			Menu ID
	 * @param {tmaModule}	module			configs owner module
	 * @param {Function}	toggleHandler	optional callback for modules config toggle handling
	 */
	this.addConfigItem = function( f, menuid, module, toggleHandler ) {

		if (menus[menuid]) {

			f.stop();

			menus[menuid].addConfigItem(module, toggleHandler);
		}

	};

}

/**
 * menu instance
 *
 * @param {String}	menuId
 * @returns {Menu}
 */
function Menu(menuId) {

	var self = this;

	var currentModule = null;

	this.menuId = menuId;

	// setup menu items click handler
	$('#' + menuId).on('click', 'LI.menuitem', function(e) {

		if ( !$(this).hasClass('disabled') ) {

			var module = $(this).prop('module');

			// close dropdown menu
			if ( !Modernizr.touch ) {
				// end hover
				var $groups = $(this).parents('UL[group]').addClass('close-dropdown');
				window.setTimeout(function() {
					$groups.removeClass('close-dropdown');
				}, 500);
			} else {
				// close other opened dropdows on touch devices
				$(this).siblings('.open-dropdown').removeClass('open-dropdown');
			}

			self.setPage(module);

		}

	});

	// setup group items click handler (touch devices only)
	if ( Modernizr.touch ) {
		$('#' + menuId).on('click', 'LI.has-dropdown', function(e) {

			if ( !$(this).hasClass('disabled') ) {
				if (e.target.parentNode == this) e.stopPropagation();
				$(this).toggleClass('open-dropdown');
				if ($(this).hasClass('open-dropdown')) {
					// close other opened dropdowns
					$(this).siblings('.open-dropdown').removeClass('open-dropdown');
				}
			}

		});
	}

	$(window).on('resize', function() {
		self.updatePositions();
	});

	/**
	 * adds a menu item for a page
	 * called from  modules, to add their pages
	 *
	 * @param {tmaModule}	module	the module should own these methods:
	 * 									show: callback to render the module content
	 * 									cleanup: callback to cleanup, what the module has modified outside content
	 * 									refresh: (optional) to refresh the content
	 * 								the module should own these properties:
	 * 									i18.menu_caption: caption of the menu entry
	 * 									pageGroup: (optional) if pages of the same cathegory should go in on submenu
	 * 									i18.group_caption: (optional) caption of the submenu
	 */
	this.addPage = function( module ) {

		var $compactMenu = $('#' + menuId + ' > LI.compact-menu > UL');
		var $menu = ($compactMenu.length) ? $compactMenu : $('#' + menuId);

		if (module.pageGroup) {

			var $item = $menu.find( 'LI[group="{0}"]'.format(module.pageGroup) );
			if ($item.length) {

				var $submenu = $menu.find( 'ul[group="{0}"]'.format(module.pageGroup) );
				if (!$submenu.length) {

					// no submenu, but an item with this group already exists
					// => create a submenu and insert existing item

					$submenu = $(
							('<li class="has-dropdown" id="group-{group}"><a>{caption}</a><ul group="{group}" />')
							.format({
								group:		module.pageGroup,
								caption:	$item.prop('module').i18.group_caption,
							})
					).insertAfter($item);

					$item.css({
						left: '',
						zIndex: '',
					});

					// move $item into submenu
					$menu = $submenu.find('UL').append($item);

				} else {

					// submenu exists
					$menu = $submenu;

				}

			}
		}

		var $li = $(
				('<li id="{itemid}" class="menuitem" {group}><a href="#{itemid}">{caption}</a></li>')
				.format({
					itemid:		module.name,
					caption:	module.i18.menu_caption,
					group:		(module.pageGroup) ? 'group="{0}"'.format(module.pageGroup) : '',
				})
		)
		.prop('module', module)
		.appendTo($menu);

		if (module.refresh) {
			$li.find('A').addClass('refreshable');
		}

		this.updatePositions();

	};

	var compactTreshold = 0;

	this.updatePositions = function() {

		var COMPRESS_LIMIT = 10;

		var $menu = $('#' + menuId);
		var $compactMenu = $menu.find('> LI.compact-menu');
		var compress = 0;
		var $li;

		if ( ($compactMenu.length > 0) && ($menu.width() > compactTreshold) ) {

			// switch back to normal mode

			$li = $compactMenu.find('> UL > LI').appendTo($menu);

			var $config = $menu.find('> LI.config').css({
				right: 	'',
				left:	0,
			});
			$config.insertAfter($li.filter('#' + $config.attr('insert')));

			$compactMenu.remove();

			$compactMenu = $();

		}

		if ($compactMenu.length === 0) {

			// normal mode: calculate required width and neccessary compression value

			$li = $menu.find('> LI');

			var zi = $li.length + 1;
			var width = 0;
			$li.each(function() {
				width += $(this).outerWidth();
			});

			var maxWidth = $menu.width();
			if ((width > maxWidth) && ($li.length > 0)) {
				compress = parseInt(0.99 + (width - maxWidth) / $li.length);
			}

			if (compress > COMPRESS_LIMIT) {

				// store current calculated width as treshold to switch back to normal mode
				compactTreshold = width;
				//console.log('compress: width = ' + width + '; maxWidth = ' + maxWidth + '; compress = ' + compress + ' x ' + $li.length);

				// switch to compact mode

				$compactMenu = $(
						'<li class="compact-menu has-dropdown">' +
							'<a><i class="fa fa-bars" /></a>' +
							'<ul group="' + menuId + '" />' +
						'</li>'
				);

				// move all items, except brand, one level deeper
				$li.filter(':not(.brand)').css({
					left: '',
					zIndex: '',
				})
				.appendTo($compactMenu.find('UL'));

				$menu.append($compactMenu);

			}

		}

		if ($compactMenu.length > 0) {

			// compact mode: move config item to top level, if existing
			$compactMenu.find('LI.config').insertBefore($compactMenu).css({
				left: '',
				right: parseInt($compactMenu.css('right')) + $compactMenu.outerWidth(),
			});

		} else {

			// normal mode: calculate positions of top level items

			$li = $menu.find('> LI');
			var _left = 0;
			$li.each(function() {

				$(this).css({
					zIndex: zi--,
					left:	_left,
				});

				_left += $(this).outerWidth() - compress;

			});

			if ($li.length === 0) {
				tma.log('normal mode: calculate positions of 0 top level items');
			} else if (_left === 0) {
				tma.log('normal mode: calculate positions of top level items: _left=0');
				//window.setTimeout(function() {
				//	self.updatePositions();
				//}, 500);
			}

		}

	};


	/**
	 * switch to, or refresh page
	 *
	 *  @param {tmaModule}	module
	 */
	this.setPage = function(module) {

		//console.log('setPage: ' + module.name);
		if (
				(module.manifest.type) &&
				(module.manifest.type == 'page')
		) {

			if( currentModule != module ) {

				// switch to new module

				// first cleanup prev. active module
				self.cleanupActiveModule();

				// switch module
				currentModule = module;

				// set selected menu item
				$('#' + menuId + ' LI#' + module.name).addClass('sel').parents('LI').addClass('sel');

				// display modules content
				if (currentModule.show) {
					// call show with modules scope
					currentModule.show.call(currentModule);
				}

			} else {

				// refresh modules content
				if (currentModule.refresh) {
					// call refresh with modules scope
					currentModule.refresh.call(currentModule);
				}

			}
		}

		this.updatePositions();

	};

	/**
	 * removes a menu item
	 * called from modules, to cleanup their config item
	 *
	 * @param {String}		itemId			id of the menu item to remove
	 */
	this.removeItem = function( itemId ) {

		$('#' + menuId + ' LI#' + itemId).remove();

	};

	/**
	 * call the cleanup method of the prev. active module
	 */
	this.cleanupActiveModule = function() {

		if ( (currentModule) && (currentModule.cleanup) ) {
			// call cleanup with modules scope
			currentModule.cleanup.call(currentModule);
		}

		// clear config
		$('#config').empty().hide();

		// clear content
		$('#head').append( $('DIV#content DIV.sticky-content') );
		$('.sticky-content').empty();
		$('#content').empty();

		// clear timestamp
		$('#timestamp').empty();

		// remove menu item
		$('#' + menuId + ' LI.config').remove();

		// reset selected menu item
		$('#' + menuId + ' LI.sel').removeClass('sel');

	};

	this.enableItem = function(itemId) {
		$('#' + menuId + ' LI#' + itemId).removeClass('disabled');
	};

	this.disableItem = function(itemId) {
		$('#' + menuId + ' LI#' + itemId).addClass('disabled');
	};

	this.showItem = function(itemId) {
		$('#' + menuId + ' LI#' + itemId).show();
	};

	this.hideItem = function(itemId) {
		$('#' + menuId + ' LI#' + itemId).hide();
	};

	this.addConfigItem = function(module, toggleHandler) {

		var insertAfterId = module.name;

		var $groupItem = $('#' + menuId + ' LI#' + insertAfterId).parents('UL[group]').last().parent();
		if ($groupItem.length && $groupItem.attr('id')) {
			insertAfterId = $groupItem.attr('id');
		}

		var clickHandler = toggleHandler || self.toggleConfig;

		$('<li id="' + module.name + '-config" class="config" insert="' + insertAfterId + '"><a><i class="fa fa-gear" /></a></li>')
		.insertAfter($('#' + menuId + ' LI#' + insertAfterId))
		.click( function(e) {

			if ( !$(this).hasClass('disabled') ) {

				clickHandler.call(null, e);

			}

		} );

		this.updatePositions();

	};

	/**
	 * (de)activate config
	 */
	this.toggleConfig = function() {

		if ($('DIV#config').is(':visible')) {

			$('DIV#config').slideUp(function() {
				$('#head').append( $('DIV#content DIV.sticky-content') );

			});

		} else {

			$('DIV#content').prepend( $('#head DIV.sticky-content') );
			$(document).scrollTop(0);
			$('DIV#config').slideDown();

		}

	};

	$(window).on('hashchange', function() {

		// hash or startpage by prefrence?
		var hash =
			window.location.hash.replace(/^#/, '') ||
			tma.localSetting('startpage', 'configpage');
		//console.log('hashchange: ' + ((currentModule) ? currentModule.name : 'x') + ' -> ' + hash);

		if (
				(currentModule) &&
				(hash != currentModule.name)
		) {
			var moduleCtx = tma.fn.getModuleData( hash );
			if (
					(moduleCtx) &&
					!$.isEmptyObject(moduleCtx.module)
			){
				self.setPage(moduleCtx.module);
			}
		}

	});

}
