
	/**
	 * creates a styled channel selector with a native select element in background
	 *
	 * @param {Array}	chBlacklist		optional array of channel-IDs, that should not appear as option
	 * @returns {jQueryObject}
	 */
	function createChannelSelector(chBlacklist) {

		/**
		 * get the previous an next channel object of a given channelid
		 * prev and next behaves circular
		 *
		 * @param {String}	channelid	(logoid)
		 * @param {String}	favGroup	if empty, return all channels
		 * 								virtual favGroups can be __analogue, __tv, __radio
		 * @param {Array}	chBlacklist		optional array of channel-IDs, that should not appear as option
		 * @returns {Object} {prev, next}
		 */
		function prevNextChannel(channelid, favGroup, chBlacklist) {

			var result = {
					prev: null,
					next: null
			};
			var channels = tma.getChannels(favGroup);

			if (channels.length) {
				var found = false;

				// preset prev with last channel, that is not in blacklist
				// (circular, if channelid is first channel)
				var i = channels.length - 1;
				while (
						(i >= 0) &&
						($.inArray(channels[i].logoid, chBlacklist) > -1)
				) {
					i--;
				}
				if ( i >= 0 ) {
					result.prev = channels[i];
				}

				/** @param {TmaChannel} channel */
				$.each(channels, function(i, channel) {

					if ( $.inArray(channel.logoid, chBlacklist) == -1 ) {
						if (channel.logoid == channelid) {
							found = true;
						} else if (found) {
							result.next = channel;
							return false; // break
						} else {
							result.prev = channel;
						}
					}

				});

				if (!result.next) {
					// (circular)
					// set next to first channel, that is not in blacklist
					i = 0;
					while (
							(i < channels.length) &&
							($.inArray(channels[i].logoid, chBlacklist) > -1)
					) {
						i++;
					}
					if ( i < channels.length ) {
						result.next = channels[i];
					}
				}

			}

			return result;

		}

		var template =
			'<div class="chselector">' +
				'<div class="chs-prev channel-panel" />' +
				'<div class="chs-next channel-panel" />' +
				'<dl class="chselect">' +
					'<select name="channels" />' +
					'<dt class="channel-panel">' +
						'<li />' +
					'</dt>' +
					'<dd>' +
						'<div>' +
							'<div class="chs-fav">' +
								'<select name="favgroup">' +
									'<option value="">{all}</option>' +
									'<option value="__tv">{tv}</option>' +
									'<option value="__radio">{radio}</option>' +
								'</select>' +
							'</div>' +
							'<ul />' +
						'</div>' +
					'</dd>' +
				'</dl>' +
			'</div>';

		var $selector = $( template.format(tma.i18.favgroups) );
		var $chselect = $selector.find('SELECT[name="channels"]');
		var $ul = $selector.find('UL');
		var $dropdown = $selector.find('DD > DIV');

		var option_tmpl = '<option value="{logoid}">{servicename}</option>';
		var vis_option_tmpl = '<li channelid="{logoid}"><i class="channellogo logo-{logoid} haslogo-{haslogo}" /><span>{servicename}</span></li>';

		var $favselect = $selector.find('SELECT[name="favgroup"]');
		/** @param {TmaFavouritesGroup} group */
		tma.favourites.eachGroup( function(nr, group) {

			if (group.count !== 0) {
				//Only list groups with members
				$favselect.append(
						('<option value="{name}">{getName}</option>').format(group)
				);
			}

		});

		// fill internal select with all channels except blacklisted
		/** @param {TmaChannel} channel */
		$.each(tma.getChannels(), function(ix, channel) {
			if ( $.inArray(channel.logoid, chBlacklist) == -1 ) {
				$(option_tmpl.format(channel)).appendTo($chselect);
			}
		});

		var favGroup = tma.localSetting('defaultfavourites', '');
		var favChannels = tma.getChannels(favGroup);
		var selectedid = '';

		$favselect.change(function() {

			favGroup = $(this).val();
			favChannels = tma.getChannels(favGroup);

			$ul.empty();

			/** @param {TmaChannel} channel */
			$.each(favChannels, function(ix, channel) {

				if ( $.inArray(channel.logoid, chBlacklist) == -1 ) {
					$(vis_option_tmpl.format(channel)).appendTo($ul);
				}

			});
			$li = $selector.find('DD LI[channelid="' + selectedid + '"]').addClass('selected');

			scrollInView();

		})
		/** @param {jQueryEvent} e */
		.click(function(e) {
			e.stopPropagation();
		})
		// initialise
		.val(favGroup).change();

		$chselect.change(function() {

			selectedid = $(this).val();

			var channel = tma.channel(selectedid);
			if (channel) {

				setCurrentChannelsFavGroup();

				$selector.find('DT').empty().append(
						$(vis_option_tmpl.format(channel)).removeAttr('channelid')
				);

				// mark selected
				$selector.find('DD LI.selected').removeClass('selected');
				$li = $selector.find('DD LI[channelid="' + selectedid + '"]').addClass('selected');

				// fill prev & next
				var prevnext = prevNextChannel(selectedid, favGroup, chBlacklist);
				var prev_next_tmpl = '<li channelid="{logoid}"><i class="channellogo logo-{logoid} haslogo-{haslogo}" /><span class="channellogo">{servicename}</span></li>';

				if (prevnext.prev) {
					$selector.find('.chs-prev')
					.html( $(prev_next_tmpl.format(prevnext.prev)) )
					.css('visibility', 'visible');
				} else {
					$selector.find('.chs-prev').css('visibility', 'hidden');
				}

				if (prevnext.next) {
					$selector.find('.chs-next')
					.html( $(prev_next_tmpl.format(prevnext.next)) )
					.css('visibility', 'visible');
				} else {
					$selector.find('.chs-next').css('visibility', 'hidden');
				}

			}

		});

		$selector.on('click', 'DD LI', function(e) {
			$chselect.val( $(this).attr('channelid') ).change();
		});

		// prev/next channel buttons
		$selector.on('click', '> DIV LI', function(e) {
			$dropdown.slideUp();
			$chselect.val( $(this).attr('channelid') ).change();
		});

		$selector.on('click', 'DL', function(e) {
			$dropdown.slideToggle();
			if ($dropdown.is(':visible') ) {

				setCurrentChannelsFavGroup();

				// scroll in view
				scrollInView();

			}
		});

		return $selector;

		function setCurrentChannelsFavGroup(  ) {

			if (selectedid) {
				var found = false;
				$.each(favChannels, function(ix, channel) {
					if (selectedid == channel.logoid) {
						found = true;
						return false; // break
					}
				});
				if (!found) {
					// show all channels
					$favselect.val('').change();
				}

			}

		}

		function scrollInView() {

			var $li = $selector.find('DD LI.selected');
			if ($li.length) {
				$ul.scrollTop(0);
				var pos = $li.position();
				var ulHeight = $ul.height();
				var liHeight = $li.height();

				if (
						( pos.top > ulHeight - liHeight ) ||
						( pos.top < 0 )
				) {
					$ul.scrollTop( pos.top - parseInt((ulHeight - liHeight) / 2) );
				}
			}

		}

	}
